# Copyright 2018 gevent contributors. See LICENSE for details.

import gc
import unittest


from greenlet import gettrace
from greenlet import settrace

from gevent.monkey import get_original
from gevent._compat import thread_mod_name
from gevent._compat import NativeStrIO

from gevent.testing import verify
from gevent.testing.skipping import skipWithoutPSUtil

from gevent import _monitor as monitor
from gevent import config as GEVENT_CONFIG

get_ident = get_original(thread_mod_name, 'get_ident')

class MockHub(object):
    _threadpool = None
    _resolver = None

    def __init__(self):
        self.thread_ident = get_ident()
        self.exception_stream = NativeStrIO()
        self.dead = False

    def __bool__(self):
        return not self.dead

    __nonzero__ = __bool__

    def handle_error(self, *args): # pylint:disable=unused-argument
        raise # pylint:disable=misplaced-bare-raise

    @property
    def loop(self):
        return self

    def reinit(self):
        "mock loop.reinit"

class _AbstractTestPeriodicMonitoringThread(object):
    # Makes sure we don't actually spin up a new monitoring thread.

    # pylint:disable=no-member

    def setUp(self):
        super(_AbstractTestPeriodicMonitoringThread, self).setUp()
        self._orig_start_new_thread = monitor.start_new_thread
        self._orig_thread_sleep = monitor.thread_sleep
        monitor.thread_sleep = lambda _s: gc.collect() # For PyPy
        self.tid = 0xDEADBEEF
        def start_new_thread(_f, _a):
            r = self.tid
            self.tid += 1
            return r

        monitor.start_new_thread = start_new_thread
        self.hub = MockHub()
        self.pmt = monitor.PeriodicMonitoringThread(self.hub)
        self.hub.periodic_monitoring_thread = self.pmt
        self.pmt_default_funcs = self.pmt.monitoring_functions()[:]
        self.len_pmt_default_funcs = len(self.pmt_default_funcs)

    def tearDown(self):
        monitor.start_new_thread = self._orig_start_new_thread
        monitor.thread_sleep = self._orig_thread_sleep
        prev = self.pmt._greenlet_tracer.previous_trace_function
        self.pmt.kill()
        assert gettrace() is prev, (gettrace(), prev)
        settrace(None)
        super(_AbstractTestPeriodicMonitoringThread, self).tearDown()


class TestPeriodicMonitoringThread(_AbstractTestPeriodicMonitoringThread,
                                   unittest.TestCase):

    def test_constructor(self):
        self.assertEqual(0xDEADBEEF, self.pmt.monitor_thread_ident)
        self.assertEqual(gettrace(), self.pmt._greenlet_tracer)

    @skipWithoutPSUtil("Verifies the process")
    def test_get_process(self):
        proc = self.pmt._get_process()
        self.assertIsNotNone(proc)
        # Same object is returned each time.
        self.assertIs(proc, self.pmt._get_process())

    def test_hub_wref(self):
        self.assertIs(self.hub, self.pmt.hub)
        del self.hub

        gc.collect()
        self.assertIsNone(self.pmt.hub)

        # And it killed itself.
        self.assertFalse(self.pmt.should_run)
        self.assertIsNone(gettrace())


    def test_add_monitoring_function(self):

        self.assertRaises(ValueError, self.pmt.add_monitoring_function, None, 1)
        self.assertRaises(ValueError, self.pmt.add_monitoring_function, lambda: None, -1)

        def f():
            "Does nothing"

        # Add
        self.pmt.add_monitoring_function(f, 1)
        self.assertEqual(self.len_pmt_default_funcs + 1, len(self.pmt.monitoring_functions()))
        self.assertEqual(1, self.pmt.monitoring_functions()[1].period)

        # Update
        self.pmt.add_monitoring_function(f, 2)
        self.assertEqual(self.len_pmt_default_funcs + 1, len(self.pmt.monitoring_functions()))
        self.assertEqual(2, self.pmt.monitoring_functions()[1].period)

        # Remove
        self.pmt.add_monitoring_function(f, None)
        self.assertEqual(self.len_pmt_default_funcs, len(self.pmt.monitoring_functions()))

    def test_calculate_sleep_time(self):
        self.assertEqual(
            self.pmt.monitoring_functions()[0].period,
            self.pmt.calculate_sleep_time())

        # Pretend that GEVENT_CONFIG.max_blocking_time was set to 0,
        # to disable this monitor.
        self.pmt._calculated_sleep_time = 0
        self.assertEqual(
            self.pmt.inactive_sleep_time,
            self.pmt.calculate_sleep_time()
        )

        # Getting the list of monitoring functions will also
        # do this, if it looks like it has changed
        self.pmt.monitoring_functions()[0].period = -1
        self.pmt._calculated_sleep_time = 0
        self.pmt.monitoring_functions()
        self.assertEqual(
            self.pmt.monitoring_functions()[0].period,
            self.pmt.calculate_sleep_time())
        self.assertEqual(
            self.pmt.monitoring_functions()[0].period,
            self.pmt._calculated_sleep_time)

    def test_call_destroyed_hub(self):
        # Add a function that destroys the hub so we break out (eventually)
        # This clears the wref, which eventually calls kill()
        def f(_hub):
            _hub = None
            self.hub = None
            gc.collect()

        self.pmt.add_monitoring_function(f, 0.1)
        self.pmt()
        self.assertFalse(self.pmt.should_run)

    def test_call_dead_hub(self):
        # Add a function that makes the hub go false (e.g., it quit)
        # This causes the function to kill itself.
        def f(hub):
            hub.dead = True
        self.pmt.add_monitoring_function(f, 0.1)
        self.pmt()
        self.assertFalse(self.pmt.should_run)

    def test_call_SystemExit(self):
        # breaks the loop
        def f(_hub):
            raise SystemExit()

        self.pmt.add_monitoring_function(f, 0.1)
        self.pmt()

    def test_call_other_error(self):
        class MyException(Exception):
            pass

        def f(_hub):
            raise MyException()

        self.pmt.add_monitoring_function(f, 0.1)
        with self.assertRaises(MyException):
            self.pmt()

    def test_hub_reinit(self):
        import os
        from gevent.hub import reinit
        self.pmt.pid = -1
        old_tid = self.pmt.monitor_thread_ident

        reinit(self.hub)

        self.assertEqual(os.getpid(), self.pmt.pid)
        self.assertEqual(old_tid + 1, self.pmt.monitor_thread_ident)



class TestPeriodicMonitorBlocking(_AbstractTestPeriodicMonitoringThread,
                                  unittest.TestCase):

    def test_previous_trace(self):
        self.pmt.kill()
        self.assertIsNone(gettrace())

        called = []
        def f(*args):
            called.append(args)

        settrace(f)

        self.pmt = monitor.PeriodicMonitoringThread(self.hub)
        self.assertEqual(gettrace(), self.pmt._greenlet_tracer)
        self.assertIs(self.pmt._greenlet_tracer.previous_trace_function, f)

        self.pmt._greenlet_tracer('event', ('args',))

        self.assertEqual([('event', ('args',))], called)

    def test__greenlet_tracer(self):
        self.assertEqual(0, self.pmt._greenlet_tracer.greenlet_switch_counter)
        # Unknown event still counts as a switch (should it?)
        self.pmt._greenlet_tracer('unknown', None)
        self.assertEqual(1, self.pmt._greenlet_tracer.greenlet_switch_counter)
        self.assertIsNone(self.pmt._greenlet_tracer.active_greenlet)

        origin = object()
        target = object()

        self.pmt._greenlet_tracer('switch', (origin, target))
        self.assertEqual(2, self.pmt._greenlet_tracer.greenlet_switch_counter)
        self.assertIs(target, self.pmt._greenlet_tracer.active_greenlet)

        # Unknown event removes active greenlet
        self.pmt._greenlet_tracer('unknown', ())
        self.assertEqual(3, self.pmt._greenlet_tracer.greenlet_switch_counter)
        self.assertIsNone(self.pmt._greenlet_tracer.active_greenlet)

    def test_monitor_blocking(self):
        # Initially there's no active greenlet and no switches,
        # so nothing is considered blocked
        from gevent.events import subscribers
        from gevent.events import IEventLoopBlocked
        events = []
        subscribers.append(events.append)

        self.assertFalse(self.pmt.monitor_blocking(self.hub))

        # Give it an active greenlet
        origin = object()
        target = object()
        self.pmt._greenlet_tracer('switch', (origin, target))

        # We've switched, so we're not blocked
        self.assertFalse(self.pmt.monitor_blocking(self.hub))
        self.assertFalse(events)

        # Again without switching is a problem.
        self.assertTrue(self.pmt.monitor_blocking(self.hub))
        self.assertTrue(events)
        verify.verifyObject(IEventLoopBlocked, events[0])
        del events[:]

        # But we can order it not to be a problem
        self.pmt.ignore_current_greenlet_blocking()
        self.assertFalse(self.pmt.monitor_blocking(self.hub))
        self.assertFalse(events)

        # And back again
        self.pmt.monitor_current_greenlet_blocking()
        self.assertTrue(self.pmt.monitor_blocking(self.hub))

        # A bad thread_ident in the hub doesn't mess things up
        self.hub.thread_ident = -1
        self.assertTrue(self.pmt.monitor_blocking(self.hub))


class MockProcess(object):

    def __init__(self, rss):
        self.rss = rss

    def memory_full_info(self):
        return self


@skipWithoutPSUtil("Accessess memory info")
class TestPeriodicMonitorMemory(_AbstractTestPeriodicMonitoringThread,
                                unittest.TestCase):

    rss = 0

    def setUp(self):
        _AbstractTestPeriodicMonitoringThread.setUp(self)
        self._old_max = GEVENT_CONFIG.max_memory_usage
        GEVENT_CONFIG.max_memory_usage = None

        self.pmt._get_process = lambda: MockProcess(self.rss)

    def tearDown(self):
        GEVENT_CONFIG.max_memory_usage = self._old_max
        _AbstractTestPeriodicMonitoringThread.tearDown(self)

    def test_can_monitor_and_install(self):
        # We run tests with psutil installed, and we have access to our
        # process.
        self.assertTrue(self.pmt.can_monitor_memory_usage())
        # No warning, adds a function

        self.pmt.install_monitor_memory_usage()
        self.assertEqual(self.len_pmt_default_funcs + 1, len(self.pmt.monitoring_functions()))

    def test_cannot_monitor_and_install(self):
        import warnings
        self.pmt._get_process = lambda: None
        self.assertFalse(self.pmt.can_monitor_memory_usage())

        # This emits a warning, visible by default
        with warnings.catch_warnings(record=True) as ws:
            self.pmt.install_monitor_memory_usage()

        self.assertEqual(1, len(ws))
        self.assertIs(monitor.MonitorWarning, ws[0].category)

    def test_monitor_no_allowed(self):
        self.assertEqual(-1, self.pmt.monitor_memory_usage(None))

    def test_monitor_greater(self):
        from gevent import events

        self.rss = 2
        GEVENT_CONFIG.max_memory_usage = 1

        # Initial event
        event = self.pmt.monitor_memory_usage(None)
        self.assertIsInstance(event, events.MemoryUsageThresholdExceeded)
        self.assertEqual(2, event.mem_usage)
        self.assertEqual(1, event.max_allowed) # pylint:disable=no-member
        self.assertIsInstance(event.memory_info, MockProcess) # pylint:disable=no-member

        # No growth, no event
        event = self.pmt.monitor_memory_usage(None)
        self.assertIsNone(event)

        # Growth, event
        self.rss = 3
        event = self.pmt.monitor_memory_usage(None)
        self.assertIsInstance(event, events.MemoryUsageThresholdExceeded)
        self.assertEqual(3, event.mem_usage)

        # Shrinking below gets us back
        self.rss = 1
        event = self.pmt.monitor_memory_usage(None)
        self.assertIsInstance(event, events.MemoryUsageUnderThreshold)
        self.assertEqual(1, event.mem_usage)

        # coverage
        repr(event)

        # No change, no event
        event = self.pmt.monitor_memory_usage(None)
        self.assertIsNone(event)

        # Growth, event
        self.rss = 3
        event = self.pmt.monitor_memory_usage(None)
        self.assertIsInstance(event, events.MemoryUsageThresholdExceeded)
        self.assertEqual(3, event.mem_usage)


    def test_monitor_initial_below(self):
        self.rss = 1
        GEVENT_CONFIG.max_memory_usage = 10


        event = self.pmt.monitor_memory_usage(None)
        self.assertIsNone(event)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Support functions for travis
# See https://github.com/travis-ci/travis-rubies/blob/9f7962a881c55d32da7c76baefc58b89e3941d91/build.sh

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys


commands = {}

def command(func):
    commands[func.__name__] = lambda: func(*sys.argv[2:])
    return func

@command
def fold_start(name, msg):
    sys.stdout.write('travis_fold:start:')
    sys.stdout.write(name)
    sys.stdout.write(chr(0o33))
    sys.stdout.write('[33;1m')
    sys.stdout.write(msg)
    sys.stdout.write(chr(0o33))
    sys.stdout.write('[33;0m\n')

@command
def fold_end(name):
    sys.stdout.write("\ntravis_fold:end:")
    sys.stdout.write(name)
    sys.stdout.write("\r\n")


def main():
    cmd = sys.argv[1]
    commands[cmd]()


if __name__ == '__main__':
    main()



import sys
import unittest

from gevent.testing import skipOnPy2

class TestSubnormalFloatsAreNotDisabled(unittest.TestCase):

    @skipOnPy2('This test always fails on Python 2')
    def test_subnormal_is_not_zero(self):
        # Enabling the -Ofast compiler flag resulted in subnormal floats getting
        # disabled the moment when gevent was imported. This impacted libraries
        # that expect subnormal floats to be enabled.
        #
        # NOTE: This test is supposed to catch that. It doesn't seem to work perfectly, though.
        # The test passes under Python 2 on macOS no matter whether -ffast-math is given or not;
        # perhaps this is a difference in clang vs gcc? In contrast, the test on Python 2.7 always
        # *fails* on GitHub actions (in both CPython 2.7 and PyPy). We're far past the EOL of
        # Python 2.7 so I'm not going to spend much time investigating.
        __import__('gevent')

        # `sys.float_info.min` is the minimum representable positive normalized
        # float, so dividing it by two gives us a positive subnormal float,
        # as long as subnormals floats are not disabled.
        self.assertGreater(sys.float_info.min / 2, 0)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2018 gevent community
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Test environment setup.

This establishes the resources that are available for use,
which are tested with `support.is_resource_enabled`.

"""
from __future__ import absolute_import, division, print_function

# This file may be imported early, so it should take care not to import
# things it doesn't need, which means deferred imports.


def get_ALL_RESOURCES():
    "Return a fresh list of resource names."
    # RESOURCE_NAMES is the list of all known resources, including those that
    # shouldn't be enabled by default or when asking for "all" resources.
    # ALL_RESOURCES is the list of resources enabled by default or with "all" resources.

    try:
        # 3.6 and 3.7
        from test.libregrtest import ALL_RESOURCES
    except ImportError:
        # 2.7 through 3.5

        # Don't do this:
        ## from test.regrtest import ALL_RESOURCES

        # On Python 2.7 to 3.5, importing regrtest iterates
        # sys.modules and does modifications. That doesn't work well
        # when it's imported from another module at module scope.
        # Also, it makes some assumptions about module __file__ that
        # may not hold true (at least on 2.7), especially when six or
        # other module proxy objects are involved.
        # So we hardcode the list. This is from 2.7, which is a superset
        # of the defined resources through 3.5.

        ALL_RESOURCES = (
            'audio', 'curses', 'largefile', 'network', 'bsddb',
            'decimal', 'cpu', 'subprocess', 'urlfetch', 'gui',
            'xpickle'
        )

    return list(ALL_RESOURCES) + [
        # Do we test the stdlib monkey-patched?
        'gevent_monkey',
    ]


def parse_resources(resource_str=None):
    # str -> Sequence[str]

    # Parse it like libregrtest.cmdline documents:

    # -u is used to specify which special resource intensive tests to run,
    # such as those requiring large file support or network connectivity.
    # The argument is a comma-separated list of words indicating the
    # resources to test.  Currently only the following are defined:

    #     all -       Enable all special resources.
    #
    #     none -      Disable all special resources (this is the default).
    # <snip>
    #     network -   It is okay to run tests that use external network
    #                 resource, e.g. testing SSL support for sockets.
    # <snip>
    #
    #     subprocess  Run all tests for the subprocess module.
    # <snip>
    #
    # To enable all resources except one, use '-uall,-<resource>'.  For
    # example, to run all the tests except for the gui tests, give the
    # option '-uall,-gui'.

    # We make a change though: we default to 'all' resources, instead of
    # 'none'. Encountering either of those later in the string resets
    # it, for ease of working with appending to environment variables.

    if resource_str is None:
        import os
        resource_str = os.environ.get('GEVENTTEST_USE_RESOURCES')

    resources = get_ALL_RESOURCES()

    if not resource_str:
        return resources

    requested_resources = resource_str.split(',')

    for requested_resource in requested_resources:
        # empty strings are ignored; this can happen when working with
        # the environment variable if not already set:
        # ENV=$ENV,-network
        if not requested_resource:
            continue
        if requested_resource == 'all':
            resources = get_ALL_RESOURCES()
        elif requested_resource == 'none':
            resources = []
        elif requested_resource.startswith('-'):
            if requested_resource[1:] in resources:
                resources.remove(requested_resource[1:])
        else:
            # TODO: Produce a warning if it's an unknown resource?
            resources.append(requested_resource)

    return resources

def unparse_resources(resources):
    """
    Given a list of enabled resources, produce the correct environment variable
    setting to enable (only) that list.
    """
    # By default, we assume all resources are enabled, so explicitly
    # listing them here doesn't actually disable anything. To do that, we want to
    # list the ones that are disabled. This is usually shorter than starting with
    # 'none', and manually adding them back in one by one.
    #
    # 'none' must be special cased because an empty option string
    # means 'all'. Still, we're explicit about that.
    #
    # TODO: Make this produce the minimal output; sometimes 'none' and
    # adding would be shorter.

    all_resources = set(get_ALL_RESOURCES())
    enabled = set(resources)

    if enabled == all_resources:
        result = 'all'
    elif resources:
        explicitly_disabled = all_resources - enabled
        result = ''.join(sorted('-' + x for x in explicitly_disabled))
    else:
        result = 'none'
    return result


def setup_resources(resources=None):
    """
    Call either with a list of resources or a resource string.

    If ``None`` is given, get the resource string from the environment.
    """

    if isinstance(resources, str) or resources is None:
        resources = parse_resources(resources)

    from . import support
    support.use_resources = list(resources)
    support.gevent_has_setup_resources = True

    return resources

def ensure_setup_resources():
    # Call when you don't know if resources have been setup and you want to
    # get the environment variable if needed.
    # Returns an object with `is_resource_enabled`.
    from . import support
    if not support.gevent_has_setup_resources:
        setup_resources()

    return support

def exit_without_resource(resource):
    """
    Call this in standalone test modules that can't use unittest.SkipTest.

    Exits with a status of 0 if the resource isn't enabled.
    """

    if not ensure_setup_resources().is_resource_enabled(resource):
        print("Skipped: %r not enabled" % (resource,))
        import sys
        sys.exit(0)

def skip_without_resource(resource, reason=''):
    requires = 'Requires resource %r' % (resource,)
    if not reason:
        reason = requires
    else:
        reason = reason + ' (' + requires + ')'

    if not ensure_setup_resources().is_resource_enabled(resource):
        import unittest
        raise unittest.SkipTest(reason)

if __name__ == '__main__':
    print(setup_resources())
